/* Classes to remotely control Kodak DC120 Digital Camera.
 * 
 * Written by:  Chris Studholme
 * Last Update: 7-July-1999
 * Copyright:   GPL (http://www.fsf.org/copyleft/gpl.html)
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

#include "dc120.h"


#define DATA_PACKET_SIZE 1025

#define DEFAULT_RETRIES 3

#define FAILURE_DELAY 500000  /* half second */
#define SETTING_DELAY 100000  /* tenth second */
#define BAUD_DELAY    100000  /* tenth second */
#define COMMAND_WAIT      20  /* two seconds */


/* ================ global methods ================ */

unsigned char* copy_data(unsigned char* dest, const unsigned char* buf, int n) {
  while (n-->0) 
    *dest++=*buf++;
  return dest;
}
char* copy_char(char* dest, const unsigned char* buf, int n) {
  while (n-->0) 
    *dest++=(char)(*buf++);
  *dest=0;
  return dest;
}
unsigned int copy_uint16(const unsigned char* buf) {
  return (buf[0]<<8) + buf[1];
}
unsigned int copy_uint32(const unsigned char* buf) {
  return (buf[0]<<24) + (buf[1]<<16) + (buf[2]<<8) + buf[3];
}

void WriteHexData(FILE* output, const unsigned char* data, int ndata) {
  for (int i=0; i<ndata; ++i) {
    if ((i&3)==0)
      fprintf(output," ");
    fprintf(output,"%.02X",data[i]);
  }
  fprintf(output,"\n");
}


void WriteCharData(FILE* output, const char* data, int ndata) {
  fprintf(output," '");
  for (int i=0; i<ndata; ++i) 
    if ((data[i]<0x20)||(data[i]>0x7E))
      fprintf(output,"{%.02X}",data[i]);
    else
      fprintf(output,"%c",data[i]);
  fprintf(output,"'\n");
}


bool CheckZeros(const unsigned char* data, int ndata) {
  for (int i=0; i<ndata; ++i)
    if (data[i]!=0)
      return false;
  return true;
}



/* ================ class dc120_cardsettings methods ================ */

void dc120_cardsettings::init(const unsigned char* buf) {
  copy_data(data,buf,21);
  copy_char(name,buf+21,11);
  copy_data(data+21,buf+32,2);
}

void dc120_cardsettings::display(FILE* output) {
  fprintf(output,"cardname: '%s'\n",name);
  fprintf(output,"carddata:");
  WriteHexData(output,data,sizeof(data));
}


/* ================ class dc120_settings methods ================ */

void dc120_settings::init(const unsigned char* buf) {
  /* unknown data contains:
   *   first 10 bytes
   *   last 20 bytes
   *   two bytes after name
   *   bytes 17,18,22,28,31,35
   */
  copy_data(data,buf,10);
  copy_data(data+10,buf+36,20);
  copy_data(data+30,buf+120,2);
  copy_data(data+32,buf+17,2);
  copy_data(data+34,buf+22,1);
  copy_data(data+35,buf+28,1);
  copy_data(data+36,buf+31,1);
  copy_data(data+37,buf+35,1);

  //  defaultFlashSetting;

  //buf[8];
  //buf[9];
  sleepTimeout=buf[10];
  lcdTimeout=buf[11];

  cameraTime=copy_uint32(buf+12);
  timeOfReading=time(0);

  focusSetting=(buf[16]>>4)&0x0F;
  zoomSetting=buf[16]&0x0F;
  //buf[17];
  //buf[18];
  qualitySetting=buf[19];

  flashSetting=buf[20];
  exposureCompensation=buf[21];
  //buf[22];
  manualExposure=buf[23]==1;

  shutterSpeed=copy_uint32(buf+24);

  //buf[28];
  selfTimer=buf[29]==1;
  cardType=buf[30];
  //buf[31];

  dateSetting=buf[32];
  timeSetting=buf[33];
  macroUnits=buf[34];
  //buf[35];

  cardSettings.init(buf+56);

  copy_char(userName,buf+90,30);
  memoryAlbums=copy_uint16(buf+122);
  cardAlbums=copy_uint16(buf+124);
  copy_data(zeros,buf+126,130);
  valid=true;
}

void dc120_settings::display(FILE* output) {
  fprintf(output,"Camera ID and Settings:\n");

  fprintf(output,"   data:");  
  WriteHexData(output,data,sizeof(data));

  fprintf(output,"   name: '%s'\n",userName);

  cardSettings.display(output);

  if (!CheckZeros(zeros,sizeof(zeros))) {
    fprintf(output,"  zeros:");  
    WriteHexData(output,zeros,sizeof(zeros));
  }
}


/* ================ class dc120_nfiles methods ================ */

void dc120_nfiles::init(const unsigned char* buf) {
  copy_data(zeros,buf,2);
  nfiles = copy_uint16(buf+2);
  copy_data(zeros+2,buf+4,12);
}

void dc120_nfiles::display(FILE* output) {
  fprintf(output,"Number of Files: %d\n",nfiles);
  
  if (!CheckZeros(zeros,sizeof(zeros))) {
    fprintf(output,"  zeros:");  
    WriteHexData(output,zeros,sizeof(zeros));
  }
}


/* ================ class dc120_file methods ================ */

void dc120_file::init(const unsigned char* buf) {
  copy_char(name,buf,12);
  copy_data(data,buf+12,4);
  size = copy_uint32(buf+16);
}

void dc120_file::display(FILE* output) {
  fprintf(output,"   name: '%s'  size: %d  data:",name,size);
  WriteHexData(output,data,sizeof(data));
}


/* ================ class dc120_filenames methods ================ */

void dc120_filenames::init(const unsigned char* buf) {
  nfiles = copy_uint16(buf);
  delete[] file;
  file = new dc120_file[nfiles];
  for (unsigned int i=0; i<nfiles; ++i)
    file[i].init(buf+2+i*file[i].structsize);
}

void dc120_filenames::display(FILE* output) {
  fprintf(output,"Filenames: %d\n",nfiles);
  for (unsigned int i=0; i<nfiles; ++i) 
    file[i].display(output);
}


/* ================ class DC120 methods ================ */


DC120::DC120(const char* ttydevice) {
  maxretries=DEFAULT_RETRIES;

  port = open(ttydevice,O_RDWR|O_NOCTTY);

  if (port<0) {
    fprintf(stderr,"DC120::DC120, unable to open port %s\n",ttydevice);
    errorcode=-99;
    return;
  }

  termios tty_attr;
  
  tcgetattr(port,&tty_attr);
  tty_attr.c_iflag=0;
  tty_attr.c_oflag=0;
  tty_attr.c_cflag=CS8|CREAD|CLOCAL; //|CRTSCTS;
  tty_attr.c_lflag=0;
  cfmakeraw(&tty_attr);
  tty_attr.c_iflag|=IGNBRK|IGNPAR;
  tty_attr.c_cc[VTIME]=COMMAND_WAIT;
  tty_attr.c_cc[VMIN]=0;
  cfsetispeed(&tty_attr,B9600);
  cfsetospeed(&tty_attr,B9600);
  tcflush(port,TCIFLUSH);
  tcsetattr(port,TCSANOW,&tty_attr);

  // initial break to camera
  tcsendbreak(port,12);  // try for 3 seconds long
  usleep(100000);
}


DC120::~DC120() {
  if (port>=0) {
    close(port);
  }
}

/* sends command and wait for acknowledge 
 */
bool DC120::sendCommand(unsigned char command, unsigned char data1=0, 
			unsigned char data2=0, unsigned char data3=0, 
			unsigned char data4=0, unsigned char data5=0, 
			unsigned char data6=0) {
  errorcode=0;
  unsigned char cmd[8];
  cmd[0]=command;
  cmd[1]=data1;
  cmd[2]=data2;
  cmd[3]=data3;
  cmd[4]=data4;
  cmd[5]=data5;
  cmd[6]=data6;
  cmd[7]=0x1A;

  // send command
  tcflush(port,TCIOFLUSH);
  write(port,cmd,sizeof(cmd));
  tcdrain(port);

  // wait for acknowledge
  unsigned char result;
  int size = read(port,&result,1);
  
  if ((size==1)||(result==0xD1))
    return true;

  // failure
  if (size==1)
    errorcode=result;
  else
    errorcode=-1;

  // delay and flush port
  tcflush(port,TCIOFLUSH);
  usleep(FAILURE_DELAY);
  tcflush(port,TCIOFLUSH);
  
  return false;
}


bool DC120::sendACK() {
  errorcode=0;
  unsigned char cmd = 0xD2;

  // send command
  tcflush(port,TCIOFLUSH);
  write(port,&cmd,1);
  tcdrain(port);

  // wait for result
  unsigned char result;
  int size = read(port,&result,1);
  if ((size==1)&&(result==0))
    return true;

  // failure
  if (size==1)
    errorcode=result;
  else
    errorcode=-1;

  // delay and flush port
  tcflush(port,TCIOFLUSH);
  usleep(FAILURE_DELAY);
  tcflush(port,TCIOFLUSH);
  
  return false;
}


/* Not sure if this byte is acknoledged or not.
 */
bool DC120::sendQUIT() {
  errorcode=0;
  unsigned char cmd = 0xE4;

  // send command
  tcflush(port,TCIOFLUSH);
  write(port,&cmd,1);
  tcdrain(port);

  // wait for result
  unsigned char result;
  int size = read(port,&result,1);
  if ((size==1)&&(result==0))
    return true;

  // failure
  if (size==1)
    errorcode=result;
  else
    errorcode=-1;

  // delay and flush port
  tcflush(port,TCIOFLUSH);
  usleep(FAILURE_DELAY);
  tcflush(port,TCIOFLUSH);
  
  return false;
}


int DC120::getData(void* buffer, int nbuffer) {
  unsigned char* buf = (unsigned char*)buffer;
  int size=0;
  int psize;
  while ((size<nbuffer)&&((psize=read(port,buf,nbuffer-size))>0)) {
    size+=psize;
    buf+=psize;
  }
  return size;
}


int DC120::getDataNext(void* buffer, int nbuffer) {
  errorcode=0;
  unsigned char cmd = 0xD2;

  // send command
  tcflush(port,TCIOFLUSH);
  write(port,&cmd,1);
  tcdrain(port);

  return getData(buffer,nbuffer);
}


int DC120::getDataAgain(void* buffer, int nbuffer) {
  errorcode=0;
  unsigned char cmd = 0xE3;

  // send command
  tcflush(port,TCIOFLUSH);
  write(port,&cmd,1);
  tcdrain(port);

  return getData(buffer,nbuffer);
}


int DC120::getDataWithRetries(void *buffer, int nbuffer, bool ack=true) {
  int tries = maxretries+1;
  int size;
  do {
    size = (tries==maxretries+1 ? getData(buffer,nbuffer) :
	    getDataAgain(buffer,nbuffer));

    // no data
    if (size<=0) {
      errorcode=-1;       // no data from camera
      return false;
    }
    
    // data is ok
    if ((size==nbuffer)&&(verifyChecksum(buffer,nbuffer))) {
      if (tries!=maxretries+1)
	errorcode=-102;   // retries required
      if ((ack)&&(!sendACK()))
	errorcode=-101;   // ack failed
      return true;
    }

  } while (--tries>0);

  // failure
  sendQUIT();
  if (size!=nbuffer)
    errorcode=-2;   // insufficient data
  else
    errorcode=-3;   // bad checksum
  return false;
}


int DC120::getNextDataWithRetries(void *buffer, int nbuffer, bool ack=false) {
  int tries = maxretries+1;
  int size;
  do {
    size = (tries==maxretries+1 ? getDataNext(buffer,nbuffer) :
	    getDataAgain(buffer,nbuffer));
    
    // no data
    if (size<=0) {
      errorcode=-1;       // no data from camera
      return false;
    }
    
    // data is ok
    if ((size==nbuffer)&&(verifyChecksum(buffer,nbuffer))) {
      if (tries!=maxretries+1)
	errorcode=-102;   // retries required
      if ((ack)&&(!sendACK()))
	errorcode=-101;   // ack failed
      return true;
    }

  } while (--tries>0);

  // failure
  sendQUIT();
  if (size!=nbuffer)
    errorcode=-2;   // insufficient data
  else
    errorcode=-3;   // bad checksum
  return false;
}


/* ---------------- public interface ---------------- */

bool DC120::setBaudRate(int newbaud) {
  if (port<=0)
    return false;

  unsigned char data2=0,data3=0;

  switch (newbaud) {
  case B9600:
    data2 = 0x96;
    break;
  case B19200:
    data2 = 0x19;
    data3 = 0x20;
    break;
  case B38400:
    data2 = 0x38;
    data3 = 0x40;
    break;
  case B57600:
    data2 = 0x57;
    data3 = 0x60;
    break;
  case B115200:
    data2 = 0x11;
    data3 = 0x52;
    break;
    //  case B230400:
    //    data2 = 0x23;
    //    data3 = 0x04;
    //    break;
  default:
    fprintf(stderr,"DC120::ChangeBaudRate, invalid baud rate\n");
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(0x41,0,data2,data3))
    return false;

  // change baud rate
  usleep(BAUD_DELAY);
  termios tty_attr;
  tcgetattr(port,&tty_attr);
  cfsetispeed(&tty_attr,newbaud);
  cfsetospeed(&tty_attr,newbaud);
  tcflush(port,TCIFLUSH);
  tcsetattr(port,TCSANOW,&tty_attr);
  usleep(BAUD_DELAY);

  return true;
}


/* ---------------- get/set camera settings ---------------- */

bool DC120::getSettingsFromCamera() {
  if (port<=0)
    return false;

  // send command
  if (!sendCommand(0x7f))
    return false;

  unsigned char data[settings.structsize];

  if (getDataWithRetries(data,settings.structsize)) {
    settings.init(data);
    return true;
  }
  return false;
}

bool DC120::setAlbumSetting(int album) {
  if (port<=0) return false;
  if ((album<0)||(album>8)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x49,0,album)) {
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setTimeDateDisplaySetting(int timeSetting, int dateSetting) {
  if (port<=0) return false;
  if ((timeSetting<0)||(timeSetting>1)||(dateSetting<0)||(dateSetting>2)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x33,0,dateSetting,timeSetting)) {
    settings.timeSetting=timeSetting;
    settings.dateSetting=dateSetting;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setMacroUnits(int units) {
  if (port<=0) return false;
  if ((units<0)||(units>1)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x34,0,units)) {
    settings.macroUnits=units;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setDefaultFlashSetting(int flashSetting) {
  if (port<=0) return false;
  if ((flashSetting<0)||(flashSetting>2)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x35,0,0,flashSetting)) {
    settings.defaultFlashSetting=flashSetting;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setStorageQuality(int quality) {
  if (port<=0) return false;
  if ((quality<0)||(quality>3)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x71,0,quality)) {
    settings.qualitySetting=quality;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setFlashSetting(int flashSetting) {
  if (port<=0) return false;
  if ((flashSetting<0)||(flashSetting>2)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x72,0,flashSetting)) {
    settings.flashSetting=flashSetting;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setFocusSetting(int focusSetting) {
  if (port<=0) return false;
  if ((focusSetting<0)||(focusSetting>2)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x73,0,focusSetting)) {
    settings.focusSetting=focusSetting;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setSelfTimer(bool selfTimer) {
  if (port<=0) return false;
  if (sendCommand(0x74,0,selfTimer?1:0)) {
    settings.selfTimer=selfTimer?1:0;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::getZoomSetting(int zoomSetting) {
  if (port<=0) return false;
  if ((zoomSetting<0)||(zoomSetting>6)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x78,0,zoomSetting)) {
    settings.zoomSetting=zoomSetting;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setCameraSleepTimeout(int timeout) {
  if (port<=0) return false;
  if ((timeout<60)||(timeout>120)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x82,0,timeout)) {
    settings.sleepTimeout=timeout;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setLCDSleepTimeout(int timeout) {
  if (port<=0) return false;
  if ((timeout<60)||(timeout>120)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x82,1,timeout)) {
    settings.lcdTimeout=timeout;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setManualExposure(bool manualExposure) {
  if (port<=0) return false;
  if (sendCommand(0x8B,0,manualExposure?1:0)) {
    settings.manualExposure=manualExposure?1:0;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setExposureCompensation(int exposureCompensation) {
  if (port<=0) return false;
  if ((exposureCompensation<-4)||(exposureCompensation>4)) {
    errorcode=-98;
    return false;
  }
  if (exposureCompensation<0) 
    exposureCompensation = 0x80-exposureCompensation;
  if (sendCommand(0x80,0,exposureCompensation)) {
    settings.exposureCompensation=exposureCompensation;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setShutterSpeed(int shutterSpeed) {
  if (port<=0) return false;
  if ((shutterSpeed<0x0000C8)||(shutterSpeed>0x186A00)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x81,0,0,(shutterSpeed>>16)&0xFF,
		     (shutterSpeed>>8)&0xFF,shutterSpeed&0xFF)) {
    settings.shutterSpeed=shutterSpeed;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

bool DC120::setCameraTime(time_t time) {
  if (port<=0) return false;
  if ((time<0)||(time>0x7FFFFFFE)) {
    errorcode=-98;
    return false;
  }
  if (sendCommand(0x75,0,(time>>24)&0xFF,(time>>16)&0xFF,
		     (time>>8)&0xFF,time&0xFF)) {
    settings.cameraTime=time;
    usleep(SETTING_DELAY);
    return true;
  }
  return false;
}

//bool DC120::setUserName(const char* userName);


/* ---------------- camera control ---------------- */

bool DC120::TakePicture() {
  if (port<=0) return false;
  return sendCommand(0x7C);
}


bool DC120::ErasePicture(bool card, int fileNumber) {
  if (port<=0) return false;
  if (fileNumber<0) {
    errorcode=-98;
    return false;
  }
  return sendCommand(0x7B,card?1:0,(fileNumber>>8)&0xFF,fileNumber&0xFF);
}


/* ---------------- query/download files ---------------- */

bool DC120::getNumberOfFiles(dc120_nfiles &s, bool card, int album) {
  if (port<=0)
    return false;

  if ((album<0)||(album>8)) {
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(0x45,card?1:0,0,0,album))
    return false;

  unsigned char data[s.structsize];
  if (getDataWithRetries(data,sizeof(data))) {
    s.init(data);
    return true;
  }
  return false;
}


bool DC120::getFileNames(dc120_filenames &names, bool card, int album) {
  if (port<=0)
    return false;

  if ((album<0)||(album>8)) {
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(0x4A,card?1:0,0,0,album))
    return false;

  /* depending on the number of filenames, multiple packets may need
   * to be retrieved
   */

  // get first 256 bytes of data
  unsigned char* data = new unsigned char[names.structsize];
  if (!getDataWithRetries(data,names.structsize,false)) {
    delete[] data;
    return false;
  }    

  unsigned int nnames = copy_uint16(data);
  unsigned int blocks = (2+nnames*dc120_file::structsize+255) / 256;

  // resize data if we have more than 12 filenames
  if (blocks>1) {
    unsigned char* newdata = new unsigned char[256*blocks+1];
    memcpy(newdata,data,256);
    delete[] data;
    data = newdata;
  }

  // get remaining data
  for (unsigned int i=1; i<blocks; ++i) 
    if (!getNextDataWithRetries(data+i*256,names.structsize,false)) {
      delete[] data;
      return false;
    }

  if (!sendACK())
    errorcode=-101;   // ack failed
  
  names.init(data);
  delete[] data;
  return true;
}


bool DC120::getFileData(void* data, bool card, int album, int filenumber) {
  if (port<=0)
    return false;

  if ((album<0)||(album>8)||(filenumber<0)||(filenumber>254)) {
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(card?0x64:0x54,1,0,filenumber+1,album))
    return false;

  return (getDataWithRetries(data,DATA_PACKET_SIZE,false));
}


bool DC120::getFileThumbnail(void* data, bool card, int album, int filenumber) {
  if (port<=0)
    return false;

  if ((album<0)||(album>8)||(filenumber<0)||(filenumber>254)) {
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(card?0x66:0x56,1,0,filenumber+1,album))
    return false;

  return (getDataWithRetries(data,DATA_PACKET_SIZE,false));
}


bool DC120::getFileInfo(void* data, bool card, int album, int filenumber) {
  if (port<=0)
    return false;

  if ((album<0)||(album>8)||(filenumber<0)||(filenumber>254)) {
    errorcode=-98;
    return false;
  }

  // send command
  if (!sendCommand(card?0x65:0x55,1,0,filenumber+1,album))
    return false;

  return (getDataWithRetries(data,DATA_PACKET_SIZE,false));
}


bool DC120::getNextData(void* data) {
  if (port<=0) return false;
  return (getNextDataWithRetries(data,DATA_PACKET_SIZE,false));
}

bool DC120::finishDataTransfer() {
  if (port<=0) return false;
  return sendACK();
}

bool DC120::abortDataTransfer() {
  if (port<=0) return false;
  return sendQUIT();
}
