#ifndef DC120_H
#define DC120_H

/* Classes to remotely control Kodak DC120 Digital Camera.
 * 
 * Written by:  Chris Studholme
 * Last Update: 7-July-1999
 * Copyright:   GPL (http://www.fsf.org/copyleft/gpl.html)
 */

#include <stdio.h>
#include <time.h>


#define byte unsigned char

class dc120_cardsettings {
 public:
  byte data[23];
  char name[12];
  unsigned int nfiles;

  static const int structsize=34;
  void init(const unsigned char* buf);
  void display(FILE* output);
};

class dc120_settings {
 public:
  byte data[38];
  byte zeros[130];

  char userName[31];
  time_t cameraTime;
  time_t timeOfReading;
  unsigned char dateSetting;
  unsigned char timeSetting;
  unsigned char qualitySetting;
  unsigned char flashSetting;
  unsigned char defaultFlashSetting;
  unsigned char selfTimer;
  unsigned char macroUnits;
  unsigned char focusSetting;
  unsigned char zoomSetting;
  unsigned char sleepTimeout;
  unsigned char lcdTimeout;
  unsigned int shutterSpeed;
  unsigned char cardType;
  unsigned char exposureCompensation;
  unsigned char manualExposure;
  unsigned int memoryAlbums;
  unsigned int cardAlbums;

  dc120_cardsettings cardSettings;
  
  bool valid;
  static const int structsize=257;
  dc120_settings() { valid=false; }
  void init(const unsigned char* buf);
  void display(FILE* output);
};

class dc120_nfiles {
 public:
  unsigned int nfiles;   
  byte zeros[14];

  static const int structsize=17;
  void init(const unsigned char* buf);
  void display(FILE* output);
};

class dc120_file {
 public:
  char name[13];
  byte data[4];
  unsigned int size;     

  static const int structsize=20;
  void init(const unsigned char* buf);
  void display(FILE* output);
};

class dc120_filenames {
 public:
  unsigned int nfiles;   
  dc120_file* file;

  static const int structsize=257;
  dc120_filenames() { file=0; }
  ~dc120_filenames() { delete[] file; }
  void init(const unsigned char* buf);
  void display(FILE* output);
};



class DC120 {
 protected:
  
  int port;

  /* possible error codes are:
     -1    no data from camera
     -2    insufficient data from camera
     -3    checksum error
     -4    no more data to be retrieved
     -98   invalid parameter
     -99   port not open
     -101  ack failed (but data is ok)
     -102  retries required (but data is ok)
     >= 0  unexpected data or error code from camera
  */     
  int errorcode;

  int maxretries;
  
 public:  // should be protected
  dc120_settings settings;

 public:
  DC120(const char* ttydevice);
  virtual ~DC120();

  int lastError() {
    return errorcode;
  }

  bool isOK() {
    return (port>=0);
  }

  bool setBaudRate(int baudRate);

  /* Set maximum number of retries.  Total number of tries is actually
   * one more than retries.  
   */
  void setMaxRetries(int retries) {
    maxretries=retries>0 ? retries : 0;
  }

  /* 
   * Commands to get and set camera settings.  Settings are cached by
   * getSettingsFromCamera().  All set commands happen immediately.
   */

  bool getSettingsFromCamera();

  /* The current album setting cannot be obtained from the camera. */
  bool setAlbumSetting(int album);

  bool setTimeDateDisplaySetting(int timeSetting, int dateSetting);
  int getTimeDisplaySetting() {
    return settings.timeSetting;
  }
  int getDateDisplaySetting() {
    return settings.dateSetting;
  }

  bool setMacroUnits(int units);
  int getMacroUnits() {
    return settings.macroUnits;
  }

  bool setDefaultFlashSetting(int flashSetting);
  int getDefaultFlashSetting() {
    return settings.defaultFlashSetting;
  }

  bool setStorageQuality(int quality);
  int getStorageQuality() {
    return settings.qualitySetting;
  }

  bool setFlashSetting(int flashSetting);
  int getFlashSetting() {
    return settings.flashSetting;
  }

  bool setFocusSetting(int focusSetting);
  int getFocusSetting() {
    return settings.focusSetting;
  }

  bool setSelfTimer(bool selfTimer);
  bool getSelfTimer() {
    return settings.selfTimer==1;
  }

  bool getZoomSetting(int zoomSetting);
  int getZoomSetting() {
    return settings.zoomSetting;
  }

  bool setCameraSleepTimeout(int timeout);
  int getCameraSleepTimeout() {
    return settings.sleepTimeout;
  }

  bool setLCDSleepTimeout(int timeout);
  int getLCDSleepTimeout() {
    return settings.lcdTimeout;
  }

  bool setManualExposure(bool manualExposure);
  bool getManualExposure() {
    return settings.manualExposure==1;
  }

  bool setShutterSpeed(int shutterSpeed);
  int getShutterSpeed() {
    return settings.shutterSpeed;
  }

  /* values are from -4 to +4 */
  bool setExposureCompensation(int exposureCompensation);
  int getExposureCompensation() {
    return settings.exposureCompensation<0x80 ? settings.exposureCompensation :
      0x80-(int)settings.exposureCompensation;
  }

  /* these two methods are an approximation, for now */
  static inline time_t cameraTimeToUnixTime(time_t time) {
    return (time>>2)+0x2D246C00;
  }
  static inline time_t unixTimeToCameraTime(time_t time) {
    return (time-0x2D246C00)<<2;
  }

  bool setCameraTime(time_t time);
  time_t getCameraTime() {
    return settings.cameraTime;
  }
  time_t getTimeSettingsRetrieved() {
    return settings.timeOfReading;
  }
  time_t getTimeDifference() {
    return cameraTimeToUnixTime(settings.cameraTime)-settings.timeOfReading;
  }

  //  bool setUserName(const char* userName);
  const char* getUserName() {
    return settings.userName;
  }


  /*
   * Commands to control the camera.
   */  

  bool TakePicture();
  bool ErasePicture(bool card, int fileNumber);


  /*
   * Commands to access files in camera or card memory.
   */  

  bool getNumberOfFiles(dc120_nfiles &nfiles, bool card, int album);
  bool getFileNames(dc120_filenames &filenames, bool card, int album);
  //  bool getAlbumNames(dc120_albums &albums);
  //  bool setAlbumNames(dc120_albums &albums);
  

  /* data is retrieved 1024 bytes at a time, buffer must have room
   * for 1025 bytes */
  bool getFileData(void* buf, bool card, int album, int fileNumber);
  bool getFileInfo(void* buf, bool card, int album, int fileNumber);
  bool getFileThumbnail(void* buf, bool card, int album, int fileNumber);
  bool getNextData(void* buf);
  bool finishDataTransfer();
  bool abortDataTransfer();


 private:

  static inline bool verifyChecksum(const void* data, int ndata) {
    unsigned char sum=0;
    unsigned char* bytes = (unsigned char*)data;
    while (ndata>0) {
      sum ^= *bytes++;
      --ndata;
    }
    return (sum==0);
  }

  bool sendCommand(unsigned char cmd, unsigned char data1=0, 
		   unsigned char data2=0, unsigned char data3=0, 
		   unsigned char data4=0, unsigned char data5=0, 
		   unsigned char data6=0);

  bool sendACK();
  bool sendQUIT();

  int getData(void* buffer, int nbuffer);
  int getDataNext(void* buffer, int nbuffer);
  int getDataAgain(void* buffer, int nbuffer);
  int getDataWithRetries(void *buffer, int nbuffer, bool ack);
  int getNextDataWithRetries(void *buffer, int nbuffer, bool ack);

};



#endif
