/* Program to download all pictures off Kodak DC120 Digital Camera.
 * 
 * Written by:  Chris Studholme
 * Last Update: 7-July-1999
 * Copyright:   GPL (http://www.fsf.org/copyleft/gpl.html)
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

#include "dc120.h"

#define PORT "/dev/ttyS0"
//#define BAUD B57600
#define BAUD B115200

#define PEDANTIC_SYNC


/* ================ global methods ================ */

/*
void WriteHexData(FILE* output, const unsigned char* data, int ndata) {
  for (int i=0; i<ndata; ++i) {
    if ((i&3)==0)
      fprintf(output," ");
    fprintf(output,"%.02X",data[i]);
  }
  fprintf(output,"\n");
}
*/


bool getPicture(unsigned char* buf, int size, bool card, int album, 
		int filenumber,	DC120& dc120) {

#ifdef PEDANTIC_SYNC
  sync();
#endif

  if (!dc120.getFileData(buf,card,album,filenumber)) {
    fprintf(stderr,"failed(%d)\n",dc120.lastError());
    return false;
  }
    
  int done = 1024;
  buf+=1024;
  int count=1;
  int totalblocks=(size+1023)/1024;
  int ndots=0;
  //  int dotspace=totalblocks/20;
  bool retry=false;
  bool badack=false;

  while (done<size) {

    if (dc120.lastError()==-101)
      badack=true;
    if (dc120.lastError()==-102)
      retry=true;

    if (((count++)*20/totalblocks)>ndots) {
#ifdef PEDANTIC_SYNC
      sync();
#endif
      fprintf(stderr,badack?"x":retry?"o":".");
      badack=false;
      retry=false;
      ++ndots;
    }

    if (!dc120.getNextData(buf)) {
      fprintf(stderr,"failed(%d)\n",dc120.lastError());
      return false;
    }
    buf+=1024;
    done+=1024;

  }

  if (dc120.lastError()==-101)
    badack=true;
  if (dc120.lastError()==-102)
    retry=true;
  fprintf(stderr,badack?"x":retry?"o":".");
    
  if (!dc120.finishDataTransfer())
    fprintf(stderr,"failed(%d)\n",dc120.lastError());
  else
    fprintf(stderr," done.\n");
  return true;
}


bool getPicturesInAlbum(bool card, int album, DC120& dc120) {
  dc120_filenames filenames;
  if (dc120.getFileNames(filenames,card,album)) 
    filenames.display(stdout);
  else {
    fprintf(stderr,"dc120.getFileNames failed (%d)\n",dc120.lastError());
    return false;
  }

  // read images
  for (unsigned int i=0; i<filenames.nfiles; ++i) {
    char filename[12];
    memcpy(filename,filenames.file[i].name,8);
    filename[7]='.';
    memcpy(filename+8,filenames.file[i].name+8,3);
    filename[11]=0;
    for (int j=0; j<11; ++j)
      filename[j] = tolower(filename[j]);
    
    fprintf(stderr,"retrieving image %s (%d/%d)",filename,i+1,
	    filenames.nfiles);
    
    unsigned int size = filenames.file[i].size;
    unsigned char buf[1024*1024];
    if (getPicture(buf,size,card,album,i,dc120)) {
      FILE* image = fopen(filename,"wb");
      if (image==0) {
	fprintf(stderr,"unable to open file %s\n",filename);
	continue;
      }
      if (fwrite(buf,size,1,image)!=1)
	fprintf(stderr,"error writing to file %s\n",filename);
      fclose(image);
    }
  }
  
  return true;
}


main() {

  DC120 dc120(PORT);

  if (!dc120.isOK()) {
    fprintf(stderr,"failed to open port\n");
    return 1;
  }

  if (dc120.getSettingsFromCamera()) 
    dc120.settings.display(stdout);
  else {
    fprintf(stderr,"dc120.getSettingsFromCamera failed (%d)\n",
	    dc120.lastError());
    return 1;
  }

  if (!dc120.setBaudRate(BAUD)) {
    fprintf(stderr,"dc120.setBaudRate failed (%d)\n",dc120.lastError());
    return 1;
  }

  dc120_nfiles nfiles;
    
  // check internal memory first
  for (unsigned int album=0; album<=dc120.settings.memoryAlbums; ++album) {
    if (dc120.getNumberOfFiles(nfiles,false,album)) 
      nfiles.display(stdout);
    else {
      fprintf(stderr,"dc120.getNumberOfFiles failed (%d)\n",dc120.lastError());
      nfiles.nfiles=0;
    }
    if ((nfiles.nfiles>0)&&(!getPicturesInAlbum(false,album,dc120)))
      fprintf(stderr,"getBlockOfPictures failed\n");
  }

  // then check memory card
  for (unsigned int album=0; album<=dc120.settings.cardAlbums; ++album) {
    if (dc120.getNumberOfFiles(nfiles,true,album)) 
      nfiles.display(stdout);
    else {
      fprintf(stderr,"dc120.getNumberOfFiles failed (%d)\n",dc120.lastError());
      nfiles.nfiles=0;
    }
    if ((nfiles.nfiles>0)&&(!getPicturesInAlbum(true,album,dc120)))
      fprintf(stderr,"getBlockOfPictures failed\n");
  }

  // restore baud rate to 9600
  if (!dc120.setBaudRate(B9600)) 
    fprintf(stderr,"dc120.setBaudRate failed (%d)\n",dc120.lastError());
  fprintf(stderr,"done.\n");


  /*
  if (!dc120.setAlbumSetting(0)) 
    fprintf(stderr,"setAlbumSetting failed (%d)\n",dc120.lastError());
  if (!dc120.setDefaultFlashSetting(0)) 
    fprintf(stderr,"setDefaultFlashSetting failed (%d)\n",dc120.lastError());
  if (!dc120.getSettingsFromCamera()) 
    fprintf(stderr,"getSettingsFromCamera failed (%d)\n",dc120.lastError());
  dc120.settings.display(stdout);

  printf("album:");
  if (!dc120.setAlbumSetting(5)) 
    fprintf(stderr,"setAlbumSetting failed (%d)\n",dc120.lastError());
  sleep(1);
  if (!dc120.getSettingsFromCamera()) 
    fprintf(stderr,"getSettingsFromCamera failed (%d)\n",dc120.lastError());
  sleep(1);
  dc120.settings.display(stdout);

  printf("flash:");
  if (!dc120.setDefaultFlashSetting(2)) 
    fprintf(stderr,"setDefaultFlashSetting failed (%d)\n",dc120.lastError());
  sleep(1);
  if (!dc120.getSettingsFromCamera()) 
    fprintf(stderr,"getSettingsFromCamera failed (%d)\n",dc120.lastError());
  sleep(1);
  dc120.settings.display(stdout);

  if (!dc120.setAlbumSetting(0)) 
    fprintf(stderr,"setAlbumSetting failed (%d)\n",dc120.lastError());
  if (!dc120.setDefaultFlashSetting(0)) 
    fprintf(stderr,"setDefaultFlashSetting failed (%d)\n",dc120.lastError());
*/
  
  return 0;
}
